import asyncio
import os

from aiohttp import web


def read_template(template_filename: str) -> str:
    try:
        with open(template_filename) as template_file:
            return template_file.read()
    except IOError:
        raise FileNotFoundError(f'Can\'t find template {template_filename} in docker image')


async def run_code(request: web.Request) -> web.Response:
    body = await request.json()
    code = body['code']
    lib_version = body['lib_version']

    template_path = request.app['TEMPLATE_PATH']

    if template_path is None:
        raise FileNotFoundError(
            'Web_server lesson type requires code template file inside docker container',
        )

    is_py_trainer = template_path.endswith('.py')

    # оборачиваем код в шаблон; кодируем, если шаблон на питоне
    wrapped_code = read_template(template_path).format(
        test_code=code.encode() if is_py_trainer else code,
    )

    # импортируем тестовую либу, если шаблон на питоне
    testlib_import_code = read_template(request.app['IMPORT_SNIPPET_PATH']).format(
        testlib_version=lib_version,
        testlib_default_version=request.app['TESTLIB_DEFAULT_VERSION'],
    ) if is_py_trainer else ''

    wrapped_code = '\n'.join((testlib_import_code, wrapped_code))

    script_path = f'/app/{request.app["SCRIPT_NAME"]}'

    with open(script_path, 'w') as f:
        f.writelines(wrapped_code)

    envs = []
    with open('/etc/environment', 'r') as env:
        for line in env.readlines():
            envs.append(line.strip())

    # перед выполнением проверющего скрипта нужно перейти в entrypoint_dir,
    # так как именно туда ставятся библиотеки
    # (в частности, node_modules)
    proc = await asyncio.create_subprocess_shell(
        f'cd {request.app["ENTRYPOINT_DIR"]} && \
        {" ".join(envs)} /bin/bash -c "{request.app["INTERPRETER_NAME"]} {script_path}"',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()

    os.remove(script_path)

    return web.json_response({
        "exit_code": proc.returncode,
        "stdout": stdout.decode(),
        "stderr": stderr.decode(),
    })



import logging
from asyncio import TimeoutError, ensure_future, sleep, subprocess, wait
from dataclasses import dataclass
from enum import Enum
from typing import List, Text

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

logger = logging.getLogger(__name__)


class WaitingProtocol(Enum):
    port_up = 1
    port_responding = 2


@dataclass
class PortConfig:
    port: int
    protocol: Text = WaitingProtocol.port_up.name


class ServerWaiter(object):
    """ Waits for ports on localhost with specified protocol """
    def __init__(self, ports_to_wait: List[PortConfig], overall_timeout: int):
        self.ports_to_wait: List[PortConfig] = ports_to_wait
        self.overall_timeout: int = overall_timeout

    @staticmethod
    def get_waiter(protocol: WaitingProtocol):
        waiters = {
            WaitingProtocol.port_up: ServerWaiter._is_port_listening,
            WaitingProtocol.port_responding: ServerWaiter._is_port_responding,
        }

        try:
            waiter = waiters[protocol]
        except KeyError:
            raise NotImplementedError(f'waiter for protocol [{protocol.name}] is not implemented')

        return waiter

    @staticmethod
    async def _is_port_listening(port: int) -> bool:
        cmd = f"netstat -tulpn 2>/dev/null | grep LISTEN | grep :{port}"
        proc = await subprocess.create_subprocess_shell(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE,
        )
        stdout, _ = await proc.communicate()

        return stdout != b''

    @staticmethod
    async def _is_port_responding(port: int) -> bool:
        timeout = ClientTimeout(total=1, connect=0.5)
        url = f'http://localhost:{port}/'
        try:
            async with ClientSession(timeout=timeout) as session:
                async with session.get(url) as resp:
                    logger.info(f'port {port} is ready, resp status [{resp.status}]')
                    return True  # we're okay with any response at all
        except (ClientError, TimeoutError) as e:
            logger.info(f'port [{port}] is still not responding, error [{e}] of type [{type(e)}]')
            return False

    async def waiting_loop(self, port_conf: PortConfig) -> None:
        is_port_ready = self.get_waiter(WaitingProtocol[port_conf.protocol])
        while not await is_port_ready(port_conf.port):
            logger.info(f'{port_conf.port} is still not responding, waiting 0.25s')
            await sleep(0.25)

        return

    async def wait_for_ports(self) -> bool:
        if not len(self.ports_to_wait):
            return True

        waiting_tasks = [ensure_future(self.waiting_loop(port_conf)) for port_conf in self.ports_to_wait]
        try:
            _, pending = await wait(waiting_tasks, timeout=self.overall_timeout)
        finally:
            for coro in waiting_tasks:
                if coro.done():
                    logger.info(coro.result())
                coro.cancel()

        return len(pending) == 0



import json
import logging
import os
import resource
from asyncio import CancelledError, Queue, Task, TimeoutError, create_task
from asyncio import subprocess as sp
from asyncio import wait_for
from dataclasses import dataclass
from typing import List, Optional, Text

import psutil
from aiohttp.web import WebSocketResponse

from serverhub_agent.utils.filesystem import FileSystem
from serverhub_agent.ws.output_processor import BaseOutputProcessor
from serverhub_agent.ws.ws_broadcaster import WsBroadcaster
from serverhub_agent.ws.ws_messages import (Command, Error, ExecReturnCode,
                                            ExecStatus, Log, LogType, Result,
                                            State, TestMsgTypes,
                                            TestStatusTypes, Types)
from serverhub_agent.ws.ws_utils import no_except

logger = logging.getLogger(__name__)


@dataclass
class RunnerMsg:
    """
    Сообщения для взаимодействия с WsBroadcaster
    """
    cmd: Command
    source: WebSocketResponse


class Runner:
    """
    Класс представляющий запущенный процесс. При старте агента создается один инстанс для всех обработчиков.

    Диспатчеры сокетов направляют stdin/run/check сообщения в этот инстанс. Сообщения обрабатывается по очереди,
    гарантируя одну последовательность State/Stdin видимую клиентам.

    Все сообщения в сокеты так же отправляются из этого класса.
    """
    reset_state = State(execution_status=ExecStatus.initialized, error_msg=None)

    def __init__(
            self,
            ws_broadcaster: WsBroadcaster,
            file_system: FileSystem,
            mem_limit_mb: int,
            output_limit: int,
            output_processor_cls: BaseOutputProcessor,
    ):
        self._queue: Queue = Queue()
        self.manager: Task = create_task(self._msg_dispatcher())
        self.ws_broadcaster: WsBroadcaster = ws_broadcaster
        self.file_system: FileSystem = file_system
        self.process: Optional[sp.Process] = None
        self.process_handlers: List[Task] = []
        self.loaded_files: List[Text] = []
        self.run_id: Optional[Text] = None
        self.mem_limit_mb: int = mem_limit_mb
        self.output_limit: int = output_limit
        self.output_processor: BaseOutputProcessor = output_processor_cls()

    # общение с раннером только через этот метод (и close())
    async def send_command(self, command: Command, src: WebSocketResponse):
        await self._queue.put(RunnerMsg(command, src))

    async def stop(self):  # завершает работу раннера после обработки предыдущих сообщений, ждет 5 сек и убивает
        await self._queue.put(RunnerMsg(None, None))
        try:
            await wait_for(self.manager, 2)
        except CancelledError:
            logger.error(f'failed to stop runner properly')

    async def _msg_dispatcher(self):  # разгребает сообщения stdin/run/check по одному
        while True:
            msg: RunnerMsg = await self._queue.get()
            command = msg.cmd
            if command is None:
                break

            if command.type in [Types.CHECK,  Types.RUN]:
                await self._start_run(command)
            elif command.type is Types.LOG and command.data.level == LogType.stdin:
                await self._send_stdin(command, msg.source)
            elif command.type is Types.STOP:
                await self._reset()
            else:
                logger.error(f"runner received command of unexpected type f{command.type}")

        logger.info('runner finished receiving commands')

    @property
    def _is_running(self):
        return self.process and self.process.returncode is None

    @no_except
    async def _reset(self):
        await self._kill_running_process()
        await self.ws_broadcaster.send_state(self.reset_state)

    @no_except
    async def _start_run(self, run_cmd: Command):
        await self._kill_running_process()
        try:
            await self._put_files(run_cmd)
        except Exception as e:
            print(e)
        await self._start_sub_process(run_cmd)

    @no_except
    async def _send_stdin(self, log_cmd: Command, source: WebSocketResponse):
        if not self._is_running:
            await self.ws_broadcaster.send_error(Error(code='WS.NoProcess', message='wait for State(in progress)'))
            return

        await self.ws_broadcaster.send_command(log_cmd, exclude=source)
        self.process.stdin.write(log_cmd.data.message.encode())
        await self.process.stdin.drain()

    async def _kill_running_process(self):
        if self._is_running:
            logger.info('killing already running process')
            process = psutil.Process(self.process.pid)
            for proc in process.children(recursive=True) + [process]:
                proc.kill()
                logger.info(f'killed [{proc.pid}]')

            for handler in self.process_handlers:
                try:
                    await wait_for(handler, 0.1)
                    logger.info('one handler finished')
                except TimeoutError:
                    logger.warning('failed to wait for process handler')

    async def _put_files(self, run_cmd: Command):
        # delete old user's files
        self.file_system.remove_tmp()

        # put new files
        try:
            self.file_system.create_files(run_cmd.data.files, is_tmp=True)
        except ValueError as e:
            logger.error(f'trying to put files outside the sandbox: error [{e}]')
            await self.ws_broadcaster.send_error(Error(code='WS.BadFiles', message='bad file structure'))
            return

    def _get_preexec_fun(self):
        max_bytes = self.mem_limit_mb * 2**20

        def set_limits():
            resource.setrlimit(resource.RLIMIT_DATA, (max_bytes, max_bytes))

        return set_limits

    async def _start_sub_process(self, run_cmd: Command):
        self.run_id = run_cmd.data.run_id
        opened_fds = len(set(os.listdir(f'/proc/{os.getpid()}/fd/')))
        logger.info(f"got [{opened_fds}] opened file descriptors")

        read_pipe_number, write_pipe_number = os.pipe()
        os.set_inheritable(write_pipe_number, True)

        if run_cmd.type is Types.RUN:
            running_state = State(execution_status=ExecStatus.run_in_progress, error_msg=None)
        else:
            running_state = State(execution_status=ExecStatus.check_in_progress, error_msg=None)
        await self.ws_broadcaster.send_state(running_state)

        # start new process (with new handlers)
        env = os.environ.copy()
        env["TEST_OUTPUT_FD"] = str(write_pipe_number)
        logger.debug(f'running [{run_cmd.data.bash_command}]')
        self.process = await sp.create_subprocess_shell(
            run_cmd.data.bash_command,
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            stdin=sp.PIPE,
            preexec_fn=self._get_preexec_fun(),
            pass_fds=(write_pipe_number,),
            env=env,
            limit=self.output_limit,
        )
        os.close(write_pipe_number)  # to avoid deadlock

        self.process_handlers = [
            create_task(self._proxy_stream(LogType.stdout)),
            create_task(self._proxy_stream(LogType.stderr)),
            create_task(self._process_waiter(
                send_result=run_cmd.type is Types.CHECK,
                test_pipe_fd=read_pipe_number,
                timeout=run_cmd.data.timeout,
            )),
        ]
        logger.info(f'spawned process with pid [{self.process.pid}]')

    @staticmethod
    def _get_test_status_by_ret_code(ret_code):
        test_statuses = {
            ExecReturnCode.ok.value: TestStatusTypes.ok,
            ExecReturnCode.failed.value: TestStatusTypes.failed,
            ExecReturnCode.failed_to_run_code.value: TestStatusTypes.failed_to_run_code,
        }
        return test_statuses.get(ret_code, TestStatusTypes.failed)

    async def _send_result(self, ret_code, test_output):
        result = Result(
            message_type=TestMsgTypes.intl if test_output else TestMsgTypes.text,
            message=test_output or "",
            status=self._get_test_status_by_ret_code(ret_code=ret_code),
            run_id=self.run_id,
        )
        await self.ws_broadcaster.send_result(result)

    async def _send_timeout_result(self):
        result = Result(
            message_type=TestMsgTypes.intl,
            message={"id": "TrainerError.Timeout"},
            status=TestStatusTypes.timeout,
            run_id=self.run_id,
        )
        await self.ws_broadcaster.send_result(result)

    async def _send_end_state(self):
        # send state null
        run_end_state = State(execution_status=ExecStatus.null, error_msg=None)
        await self.ws_broadcaster.send_state(run_end_state)

    async def _process_waiter(self, send_result: bool, test_pipe_fd: int, timeout: Optional[int]):
        try:
            ret_code = await wait_for(self.process.wait(), timeout=timeout)  # Если timeout is None, ждём завершения
        except TimeoutError:
            logger.info(f'killing process [{self.process.pid}] by timeout')
            await self._send_timeout_result()
            await self._kill_running_process()
        else:
            logger.info(f'process terminated with exit code [{ret_code}]')

            # send result only in check
            if not send_result:
                return

            with os.fdopen(test_pipe_fd) as pipe:
                test_output = pipe.read()
                try:
                    test_output = json.loads(test_output)
                except (json.JSONDecodeError, ValueError):
                    logger.error(f'failed to parse [{test_output}] as json')
                await self._send_result(ret_code=ret_code, test_output=test_output)
        finally:  # this waiter is canceled on process killing, we want state to be delivered anyway
            # os.close(test_pipe_fd)
            await self._send_end_state()

    async def _proxy_stream(self, stream_type: LogType):
        stream = self.process.stdout if stream_type is LogType.stdout else self.process.stderr

        at_eof = False
        while not at_eof:
            logger.debug(f"Subprocess {stream_type.name}: waiting read")
            try:
                line = await stream.readline()
            except CancelledError:
                logger.warning(f'Subprocess {stream_type.name}: canceled on read')
                return

            at_eof = stream.at_eof()

            if not line:
                continue
            messages = await self.output_processor.get_log(line)
            try:
                for message_type, message in messages:
                    await self.ws_broadcaster.send_log(Log(
                        message_type=message_type,
                        message=message,
                        level=stream_type,
                    ))
            except CancelledError:
                logger.warning(f'Subprocess {stream_type.name}: canceled on send')
                return

        logger.debug(f"Subprocess {stream_type.name}: eof/terminated")



import logging
from asyncio import CancelledError, wait_for
from asyncio.queues import Queue
from asyncio.tasks import Task, create_task
from dataclasses import dataclass
from enum import IntEnum
from typing import List, Optional, Union

from aiohttp.web_ws import WebSocketResponse

from serverhub_agent.types.exceptions import ClosedSocket
from serverhub_agent.ws.ws_messages import (Command, Error, ExecStatus, Log,
                                            Result, State, Types)
from serverhub_agent.ws.ws_utils import no_except, send_command_to_socket

logger = logging.getLogger(__name__)


class BcMsgType(IntEnum):
    subscribe = 1
    unsubscribe = 2
    send = 3


@dataclass
class SendParams:
    cmd: Command
    exclude: Optional[WebSocketResponse]


@dataclass
class BcMsg:
    """
    Сообщения для взаимодействия с WsBroadcaster
    """
    type: BcMsgType
    data: Union[None, SendParams, WebSocketResponse]


class WsBroadcaster:
    """
    Класс буферизующий и рассылающий сообщения на время работы одного запуска кода (неидеальная альтернатива редису).

    При старте агента создается один инстанс для всех обработчиков. Новый запуск перетирает сообщения предыдущего.

    Каждый запуск начинается сообщением State(execution_status='.._in_progress') и заканчивается сообщением
    State(execution_status='null'). Их разделяет произвольное кол-во сообщений типов Log, Error, Result (последнее может
    быть только одно и присутствует только у запуска тестов).

    При установлении коннекта по новому сокету в него посылаются все накопленные сообщения (в буфере всегда есть, как
    минимум, одно сообщение State(execution_status='null'), которое заносится туда при старте).

    Для отправки любых сообщений в сокет, кроме ошибок десериализации сообщений, должен использоваться объект этого
    класса. Это верно и для сообщений типа Log(stdin). Мы их отправляем обратно в сокет т.к. хотим, чтобы
    пользователь видел свой ввод при реконекте. В случае с Log(stdin) важно отправлять сообщения в сокет до отправки их
    в stdin процесса, чтобы гарантировать, что stdout порожденный этим stdin'ом был получен клиентом после отправленного
    stdin. (Однако, мы не можем гарантировать, что stdout, порождаемый процессом без влияния полученного stdin будет
    отображен в верном порядке относительно него).

    Log(stdin) транслируется всем, в т.ч. отправителю, чтобы зафиксировать порядок сообщения Log(stdin) относительно
    сообщений типа State, в котором оно было обработано. Мы гарантируем, что ввод, пришедший после
    первого State(in_progress) был направлен первому запуску, а после второго State(in_progress) -- второму.

    Таким образом после получения State(in_progress) можно смело очищать консоль -- весь ввод, доставленый в новый
    запуск (и возможно произведенный в параллельной сессии), будет доставлен в сокет после полученного стейта.
    """
    states_clearing_buffer = [ExecStatus.initialized, ExecStatus.run_in_progress, ExecStatus.check_in_progress]
    preinit_state = Command(Types.STATE, State(execution_status=ExecStatus.new, error_msg=None))

    def __init__(self, output_limit: int):
        self._all_ws: List[WebSocketResponse] = []
        self._queue: Queue = Queue()  # очередь еще не отосланых в сокеты сообщений И запросов на подключение/отключение
        self._buffer: List[Command] = [self.preinit_state]  # буфер сообщений для отсылки подключившимся позже
        self._manager: Task = create_task(self._msg_dispatcher())
        self.output_size = 0
        self.output_limit: int = output_limit  # todo: drop after exceed

    # запрос на отправку команды во все подписанные сокеты
    async def send_command(self, command: Command, exclude: Optional[WebSocketResponse] = None):
        await self._queue.put(BcMsg(BcMsgType.send, SendParams(command, exclude)))

    async def subscribe(self, ws: WebSocketResponse):  # запрос на подписку на рассылку команд
        await self._queue.put(BcMsg(BcMsgType.subscribe, ws))

    async def unsubscribe(self, ws: WebSocketResponse):  # запрос на отписку от рассылки
        await self._queue.put(BcMsg(BcMsgType.unsubscribe, ws))

    async def stop(self):  # останавливает рассылку сообщений, после отправки пришедших ранее, ждет 5 секунд и убивает
        await self._queue.put(None)
        try:
            await wait_for(self._manager, 2)
        except CancelledError:
            logger.error(f'failed to stop broadcaster properly')

    async def send_state(self, state: State):
        cmd = Command(type=Types.STATE, data=state)
        await self.send_command(cmd)

    async def send_error(self, error: Error):
        cmd = Command(type=Types.ERROR, data=error)
        await self.send_command(cmd)

    async def send_log(self, log: Log):
        cmd = Command(type=Types.LOG, data=log)
        await self.send_command(cmd)

    async def send_result(self, result: Result):
        cmd = Command(type=Types.RESULT, data=result)
        await self.send_command(cmd)

    def _update_buffer(self, cmd: Command):
        """
        Добавляет пришедшую команду в буфер. Команды STATE(initialized),STATE(check_in_progress),STATE(run_in_progress)
        требуют очистки предыдущих сообщений в буфере,
        т.к. данная команды отсылаются при новом запуске и после завершения инициализации контейнера.
        """
        if cmd.type is Types.STATE and cmd.data.execution_status in self.states_clearing_buffer:
            self.output_size = 0
            self._buffer = []

        if cmd.type is Types.LOG:
            self.output_size += len(cmd.data.message)
        if cmd.type is Types.LOG and self.output_size > self.output_limit:
            return

        self._buffer.append(cmd)

    async def _msg_dispatcher(self):
        """
        Разгребает очередь. В ней сообщения трех типов -- запрос на подписку/отписку сокета, запрос на отправку
        команды во все подписанные сокеты.
        """
        while True:
            msg: BcMsg = await self._queue.get()
            if msg is None:
                break

            if msg.type is BcMsgType.subscribe and isinstance(msg.data, WebSocketResponse):
                await self._subscribe(msg)
            elif msg.type is BcMsgType.unsubscribe and isinstance(msg.data, WebSocketResponse):
                await self._unsubscribe(msg)
            elif msg.type is BcMsgType.send and isinstance(msg.data, SendParams):
                await self._send(msg)
            else:
                logger.error(f'broadcaster: bad msg [{msg}]')

        logger.info('broadcaster finished receiving commands')

    @no_except
    async def _subscribe(self, msg: BcMsg):
        ws: WebSocketResponse = msg.data
        try:
            for command in self._buffer:
                await send_command_to_socket(command, ws)
        except ClosedSocket:
            logger.error('socket was closed just after subscription')
        else:
            self._all_ws.append(ws)

    @no_except
    async def _unsubscribe(self, msg: BcMsg):
        ws: WebSocketResponse = msg.data
        try:
            self._all_ws.remove(ws)
        except ValueError:
            logger.warning('socket has been unsubscribed already')

    @no_except
    async def _send(self, msg: BcMsg):
        cmd: Command = msg.data.cmd
        self._update_buffer(cmd)
        await self._send_to_all(msg.data.cmd, msg.data.exclude)

    async def _send_to_all(self, command: Command, exclude: Optional[WebSocketResponse] = None):
        for ws in self._all_ws:
            if ws is exclude:
                logger.debug(f"skipped source on stdin send")
                continue

            try:
                await send_command_to_socket(command, ws)
            except ClosedSocket:
                logger.error(f'socket closed on broadcasting, unsubscribing it')
                self._all_ws.remove(ws)



import os
from typing import TypeVar


class EnvType:
    """
    Available env types.
    """
    local = 'local'


class CodeRunnerType:
    """
    Available code runners.
    """
    subprocess = 'subprocess'
    swarm = 'swarm'


V = TypeVar('V')


def get_setting(name: str, default: V = None, cast: type = None, required: bool = False):
    """
    Get settings from env vars.
    """
    value: V = os.getenv(name)

    if value is None:
        if required:
            raise EnvironmentError(f'Required setting "{name}" not found in environment')
        return default

    if default is not None and cast is None:
        cast = type(default)
    elif cast is None:
        cast = str

    if cast is bool:
        truth_values = ('yes', 'true', '1')
        false_values = ('no', 'false', '0')
        if value.lower() not in truth_values and value.lower() not in false_values:
            raise EnvironmentError(f'For setting "{name}" use one of {truth_values} or {false_values}')
        return value.lower() in truth_values

    return cast(value)



import json
import logging
from typing import Any, List, Tuple

from serverhub_agent.types.jupyter import JupyterOutputs
from serverhub_agent.ws.ws_messages import TestMsgTypes

logger = logging.getLogger(__name__)


class BaseOutputProcessor(object):
    async def get_log(self, line: bytes) -> List[Tuple[TestMsgTypes, Any]]:
        if type(line) == bytes:
            line = line.decode()

        return [(TestMsgTypes.text, line)]


class JupyterOutputProcessor(BaseOutputProcessor):
    def __init__(self):
        # TODO: нет парсеров для output_type: error и execute_result
        self.jupyter_type_parsers: dict = {
            "stream": self.parse_stream,
            "display_data": self.parse_display,
        }

        self.jupyter_outputs_schema = JupyterOutputs.Schema()

    def parse_stream(self, output: dict) -> (TestMsgTypes, Any):
        return TestMsgTypes.text, output.get('text', '')

    def parse_display(self, output: dict) -> (TestMsgTypes, Any):
        key = None
        data = output.get('data', {})
        for key in data:
            if key.startswith('image'):
                break
            else:
                key = None

        if not key:
            return None, None

        encoded_image = data.get(key, None)
        return TestMsgTypes.image, f'data:{key};base64, {encoded_image}'

    async def get_log(self, line: bytes) -> List[Tuple[TestMsgTypes, Any]]:
        if type(line) == bytes:
            line = line.decode()

        try:
            cell_outputs = json.loads(line)
        except (json.JSONDecodeError, ValueError):
            logger.debug(f"Output is not a json [{line}]")
            return [(TestMsgTypes.text, line)]

        errors = self.jupyter_outputs_schema.validate({"outputs": cell_outputs})
        if errors:
            logger.debug(f"Output is not cell_outputs [{line}]")
            return [(TestMsgTypes.text, line)]

        result = []
        for cell_output in cell_outputs:
            if type(cell_output) is not dict:
                result.append((TestMsgTypes.text, cell_output))
                continue

            parser = self.jupyter_type_parsers.get(cell_output.get('output_type', None), None)
            if parser is None:
                continue

            o_type, o_data = parser(cell_output)
            if not o_data:
                continue

            result.append((o_type, o_data))

        return result



from ._tools import get_setting

FS_EXPOSED_DIR = get_setting('FS_EXPOSED_DIR', default='/app/')
ENTRYPOINT_DIR = get_setting('ENTRYPOINT_DIR', default='/.utils/')
REQUEST_MAX_BODY_SIZE_MB = get_setting('REQUEST_MAX_BODY_SIZE_MB', default=30)
STARTUP_TIMEOUT = get_setting('STARTUP_TIMEOUT', default=50)
PORTS_TO_WAIT = get_setting('PORTS_TO_WAIT', default='3000').split(',')
PORTS_TO_WAIT_RESPONDING = get_setting('PORTS_TO_WAIT_RESPONDING', default='').split(',')
PROC_MEM_LIMIT_MB = get_setting('PROC_MEM_LIMIT_MB', default=512)
WAIT_ON_PUT = get_setting('WAIT_ON_PUT', default=True)
INTERPRETER_NAME = get_setting('INTERPRETER_NAME', default='python')
SCRIPT_NAME = get_setting('SCRIPT_NAME', default='__test.py')
ENTRYPOINT_PATH = get_setting('ENTRYPOINT_PATH', default='/.utils/entrypoint.sh')
ENTRYPOINT_LOGS_PATH = get_setting('ENTRYPOINT_LOGS_PATH', default='/.utils/logs')



from enum import Enum
from typing import Any, List, Optional, Text

from marshmallow import ValidationError, post_load, pre_dump
from marshmallow_dataclass import dataclass

from serverhub_agent.types.common import ObjectWithSchema
from serverhub_agent.types.files import File


class TestMsgTypes(Enum):
    text = 'text'
    intl = 'intl'
    markdown = 'markdown'
    image = 'image'


class TestStatusTypes(Enum):
    ok = 'ok'
    failed = 'failed'
    failed_to_run_code = 'failed_to_run_code'
    error = 'error'
    timeout = 'timeout'
    pending = 'pending'
    null = 'null'


class ExecReturnCode(Enum):
    ok = 0
    failed = 1
    failed_to_run_code = 2


# represents the status of the running container
class ExecStatus(Enum):
    new = 'new'                                 # not initialized
    initialized = 'initialized'                 # initialization finished (signals that container started serving)
    null = 'null'                               # initialized and idle (signals that run terminated)
    check_in_progress = 'check_in_progress'     # initialized and running check
    run_in_progress = 'run_in_progress'         # initialized and running code


# types of command which can be used in WS
# communication protocol
class Types(Enum):
    RUN = 'RUN'                     # run (code)
    ERROR = 'ERROR'
    LOG = 'LOG'
    RESULT = 'RESULT'
    STATE = 'STATE'
    CHECK = 'CHECK'                 # run check
    FILE_UPLOAD = 'FILE_UPLOAD'
    INIT_REQUEST = 'INIT_REQUEST'   # request to start initialization
    START = 'START'                 # end of initialization, start serving
    RESET = 'RESET'                 # reset msg buffer and cancel running container
    STOP = 'STOP'                   # stop execution (run or check command)


class LogType(Enum):
    stdout = "stdout"
    stderr = "stderr"
    stdin = "stdin"


command_registry = {}


def register(klass_type):
    def wrapper(klass):
        if klass_type in command_registry:
            raise ValueError(
                f'class {klass_type} already registered with {command_registry[klass_type]}',
            )
        command_registry[klass_type] = klass
        return klass
    return wrapper


@dataclass
class Command(ObjectWithSchema):
    type: Types
    data: Optional[Any]

    @post_load()
    def parse_data(self, command, **kwargs):
        data = command['data']  # bug here
        data = {} if data is None else data
        try:
            payload_class = command_registry[command['type']]
        except KeyError as e:
            raise ValidationError(f'Not registered type {e}')
        data = payload_class.Schema().load(data)
        command['data'] = data
        return command

    @pre_dump()
    def dump_data(self, command, **kwargs):
        if command.data is not None and not isinstance(command.data, dict):
            data = command.data.Schema().dump(command.data)
            command.data = data
        return command


# Error type like Intl messages
@register(Types.ERROR)
@dataclass
class Error(ObjectWithSchema):
    message: str
    code: str

# Type for code execution state, for now
# wee are are intersted in execution_status and
# error_msg on ExecStatus.error status
@register(Types.STATE)
@dataclass
class State(ObjectWithSchema):
    execution_status: ExecStatus
    error_msg: Optional[Error]


# Logging command data, like from slow_ds trainer
@register(Types.LOG)
@dataclass
class Log(ObjectWithSchema):
    message_type: TestMsgTypes
    message: Any
    level: LogType


@register(Types.RESULT)
@dataclass
class Result(ObjectWithSchema):
    message_type: TestMsgTypes
    message: Any
    status: TestStatusTypes
    run_id: Optional[Text]


# разделение на Check/Run нужно пока мы кешируем логи в контейнере, т.к. фронт должен знать, что запускалось
@register(Types.CHECK)
@register(Types.RUN)
@dataclass
class Run(ObjectWithSchema):
    bash_command: Text
    files: Optional[List[File]]
    run_id: Optional[Text]
    timeout: Optional[int]


register(Types.FILE_UPLOAD)(File)


@register(Types.STOP)
@register(Types.RESET)
@register(Types.START)
@dataclass
class NoData(ObjectWithSchema):
    ...


@register(Types.INIT_REQUEST)
@dataclass
class InitRequest(ObjectWithSchema):
    request_id: Text



import asyncio
import os
import shutil

from aiohttp import web


async def run_test(request: web.Request) -> web.Response:
    body = await request.json()
    """
        body = {
            "files":[
                {
                    content: Text,
                    name: Text,
                }
            ]
        }
    """
    files = body['files']
    tests_path = f'/tests'

    for test_file in files:
        test_file_content = test_file['content']
        test_file_path =f"{tests_path}/{test_file['name']}"
        test_file_dirname = os.path.dirname(test_file_path)
        if not os.path.exists(test_file_dirname):
            os.makedirs(test_file_dirname, exist_ok=True)
        with open(test_file_path, 'w') as f:
            f.writelines(test_file_content)

    envs = []
    with open('/etc/environment', 'r') as env:
        for line in env.readlines():
            envs.append(line.strip())

    proc = await asyncio.create_subprocess_shell(
        f'cd {tests_path} && {" ".join(envs)} /bin/bash {tests_path}/run.sh',
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout, stderr = await proc.communicate()
    shutil.rmtree(tests_path)

    return web.json_response({
        "exit_code": proc.returncode,
        "stdout": stdout.decode(),
        "stderr": stderr.decode(),
    })



import logging
from asyncio import CancelledError

from aiohttp.web import WebSocketResponse

from serverhub_agent.types.exceptions import ClosedSocket
from serverhub_agent.ws.ws_messages import Command, Error, Types

logger = logging.getLogger(__name__)


async def send_command_to_socket(command: Command, socket: WebSocketResponse):
    try:
        await socket.send_json(Command.Schema().dump(command))
    except Exception as e:
        logger.info(f'on write to socket error {e}')
        raise ClosedSocket()


async def send_error_to_socket(error: Error, socket: WebSocketResponse):
    cmd = Command(type=Types.ERROR, data=error)
    await send_command_to_socket(cmd, socket)


def no_except(fun):
    async def wrapper(*args, **kwargs):
        try:
            return await fun(*args, **kwargs)
        except CancelledError as e:
            raise e
        except Exception as e:
            logger.error(f'fun [{fun.__name__}] caught an unexpected exception of type [{type(e)}] data [{str(e)}]')
            return None

    return wrapper



from serverhub_agent.settings import _env_variables
from serverhub_agent.ws.output_processor import BaseOutputProcessor


def init_settings_from_env(app):
    app['FS_EXPOSED_DIR'] = _env_variables.FS_EXPOSED_DIR
    app['ENTRYPOINT_DIR'] = _env_variables.ENTRYPOINT_DIR
    app['REQUEST_MAX_BODY_SIZE_MB'] = _env_variables.REQUEST_MAX_BODY_SIZE_MB
    app['STARTUP_TIMEOUT'] = _env_variables.STARTUP_TIMEOUT
    app['PORTS_TO_WAIT'] = _env_variables.PORTS_TO_WAIT
    app['PORTS_TO_WAIT_RESPONDING'] = _env_variables.PORTS_TO_WAIT_RESPONDING
    app['PROC_MEM_LIMIT_MB'] = _env_variables.PROC_MEM_LIMIT_MB
    app['WAIT_ON_PUT'] = _env_variables.WAIT_ON_PUT
    app['INTERPRETER_NAME'] = _env_variables.INTERPRETER_NAME
    app['SCRIPT_NAME'] = _env_variables.SCRIPT_NAME
    app['ENTRYPOINT_PATH'] = _env_variables.ENTRYPOINT_PATH
    app['ENTRYPOINT_LOGS_PATH'] = _env_variables.ENTRYPOINT_LOGS_PATH
    app['PROC_OUTPUT_LIMIT'] = 2 ** 20
    app['PROC_OUTPUT_BUFFER_SIZE'] = 2 ** 16
    app['OUTPUT_PROCESSOR_CLS'] = BaseOutputProcessor



import logging
from typing import List

from aiohttp import web

from serverhub_agent.utils.waiter import (PortConfig, ServerWaiter,
                                          WaitingProtocol)

logger = logging.getLogger(__name__)


async def wait_server_ready(request: web.Request) -> web.Response:
    ports_to_wait: List[PortConfig] = [
        PortConfig(port, WaitingProtocol.port_up.name)
        for port in request.app["PORTS_TO_WAIT"] if port
    ]
    ports_to_wait += [
        PortConfig(port, WaitingProtocol.port_responding.name)
        for port in request.app["PORTS_TO_WAIT_RESPONDING"] if port
    ]
    all_up = await ServerWaiter(ports_to_wait, request.app["STARTUP_TIMEOUT"]).wait_for_ports()
    if all_up:
        return web.json_response({"status": "ok"})
    else:
        raise web.HTTPGatewayTimeout()



import logging
from asyncio import CancelledError, Queue, Task, create_task, wait_for
from json import JSONDecodeError

from aiohttp.http_websocket import WSMessage
from aiohttp.web import Application, Request, WebSocketResponse, WSMsgType
from aiohttp.web_runner import GracefulExit
from marshmallow import ValidationError

from serverhub_agent.types.exceptions import BadMessage, ClosedSocket
from serverhub_agent.types.files import File
from serverhub_agent.utils.filesystem import FileSystem
from serverhub_agent.ws.ws_broadcaster import WsBroadcaster
from serverhub_agent.ws.ws_messages import (Command, Error, ExecStatus,
                                            LogType, State, Types)
from serverhub_agent.ws.ws_runner import Runner
from serverhub_agent.ws.ws_utils import send_error_to_socket

logger = logging.getLogger(__name__)


class Closing(Exception):
    ...


class ContainerInitializer:
    """
    Реализация протокола инициализации контейнера:
    - при подключении к неинициализированному контейнеру клиенты получают STATE(new);
    - в ответ клиенты отправляют запрос INIT_REQ(uuid) на выполнение инициализации контейнера;
    - агент броадкастит первый пришедший запрос -- его отправитель будет производить инициализацию;
    - инициализация выполняется путем загрузки файлов в контейнер последовательными командами FILE_UPLOAD;
    - инициализация завершается после получения от инициализирующего клиента команды типа START;
    - по завершении инициализации агент отправляет всем клиентам команды STATE(initialized);
    - TODO: клиент отвалился до того, как завершил инциализацию
    """
    def __init__(self, ws_broadcaster: WsBroadcaster, file_system: FileSystem):
        self._queue: Queue = Queue()
        self._manager: Task = create_task(self._msg_dispatcher())
        self._ws_broadcaster: WsBroadcaster = ws_broadcaster
        self.initializing: bool = False  # todo: add inited state, send errors on FILE_UPLOAD after init
        self.file_system: FileSystem = file_system

    async def send_command(self, command: Command):  # общение с инициализатором только через этот метод (и stop())
        await self._queue.put(command)

    async def stop(self):  # останавливает выполнение команд, после отправки пришедших ранее, ждет 5 секунд и убивает
        await self._queue.put(None)
        try:
            await wait_for(self._manager, 2)
        except CancelledError:
            logger.error(f'failed to stop initializer properly')

    async def _msg_dispatcher(self):  # разгребает сообщения INIT_REQ/FILE_UPLOAD/START по одному
        while True:
            command: Command = await self._queue.get()
            if command is None:
                break

            if command.type is Types.INIT_REQUEST:
                if not self.initializing:
                    self.initializing = True
                    await self._ws_broadcaster.send_command(command)
            elif command.type is Types.FILE_UPLOAD:
                file: File = command.data
                self.file_system.create_file(file)  # todo: do it in a subprocess
            elif command.type is Types.START:
                inited_state = Command(Types.STATE, State(execution_status=ExecStatus.initialized, error_msg=None))
                await self._ws_broadcaster.send_command(inited_state)

        logger.info('initializer finished receiving commands')


class MsgDispatcher:
    """
    Диспатчер сообщений из одного конкретного сокета, направляет их синглтон раннеру/инициализатору контейнера
    """
    def __init__(
            self,
            ws: WebSocketResponse,
            runner: Runner,
            initializer: ContainerInitializer,
    ):
        self.ws: WebSocketResponse = ws
        self.runner: Runner = runner
        self.initializer: ContainerInitializer = initializer

    def load_command(self, msg: WSMessage) -> Command:
        if msg.type != WSMsgType.TEXT:
            logger.warning(f'ignoring unknown message type [{msg.type}]')
            raise BadMessage(f'bad message type [{msg.type}]')

        try:
            command = Command.Schema().loads(msg.data)
        except (JSONDecodeError, ValidationError) as e:
            logger.warning(f'failed to decode message with error [{e}]')
            raise BadMessage(f"bad message structure [{msg.data}]")

        return command

    async def run_until_closed(self):
        try:
            async for msg in self.ws:  # stops iteration on closed socket
                try:
                    command: Command = self.load_command(msg)
                except BadMessage as e:
                    await send_error_to_socket(Error(code='WS.BadCommand', message=str(e)), self.ws)
                    continue

                if command.type in [Types.INIT_REQUEST, Types.FILE_UPLOAD, Types.START]:
                    await self.initializer.send_command(command)
                elif (
                        command.type in [Types.CHECK, Types.RUN, Types.STOP]
                        or command.type is Types.LOG and command.data.level == LogType.stdin
                ):
                    await self.runner.send_command(command, self.ws)  # ws, cause we don't want sender to receive stdin
                elif command.type is Types.RESET:
                    logger.info('received a reset command, stopping agent')
                    raise Closing
                else:
                    await send_error_to_socket(
                        Error(code='WS.BadCommand', message=f'unexpected command of type [{command.type}]'),
                        self.ws,
                    )
        except ClosedSocket:
            ...

        logger.info(f'socket was closed')


class WsServer:
    def __init__(self, app: Application):
        self.ws_broadcaster: WsBroadcaster = WsBroadcaster(app['PROC_OUTPUT_LIMIT'])
        self.initializer: ContainerInitializer = ContainerInitializer(
            self.ws_broadcaster, app['filesystem'])
        self.runner: Runner = Runner(
            self.ws_broadcaster,
            app['filesystem'],
            app['PROC_MEM_LIMIT_MB'],
            app['PROC_OUTPUT_BUFFER_SIZE'],
            app['OUTPUT_PROCESSOR_CLS'],
        )

    async def connect(self, request: Request) -> WebSocketResponse:
        ws = WebSocketResponse()
        await ws.prepare(request)
        await self.ws_broadcaster.subscribe(ws)

        try:
            dispatcher = MsgDispatcher(ws, self.runner, self.initializer)
            await dispatcher.run_until_closed()
        except Closing:
            await self.stop()
        finally:
            await self.ws_broadcaster.unsubscribe(ws)

        return ws

    async def stop(self):
        await self.initializer.stop()
        await self.runner.stop()
        await self.ws_broadcaster.stop()
        raise GracefulExit



from typing import List, Optional, Text

from marshmallow_dataclass import dataclass

from .common import ObjectWithSchema


@dataclass
class File(ObjectWithSchema):
    """
    Represents code in a file:

    {
      "language": "html",
      "content": "<h1>Hello World</h1>",
      "name": "index.html",
      "isDir": False,
      "isBin": False,
    }
    """
    content: Text
    language: Text
    name: Text
    isDir: Optional[bool]
    isBin: Optional[bool] = False


@dataclass
class Files(ObjectWithSchema):
    files: List[File]


@dataclass
class TestFile(ObjectWithSchema):
    content: str
    name: str


@dataclass
class AstTests(ObjectWithSchema):
    testlib_version: str
    precode: Optional[TestFile]
    user: TestFile
    answer: TestFile
    test: TestFile



import logging

from aiohttp import web

from serverhub_agent.routes import setup_routes
from serverhub_agent.settings import _env_variables, init_settings_from_env
from serverhub_agent.utils.filesystem import FileSystem
from serverhub_agent.ws.output_processor import BaseOutputProcessor
from serverhub_agent.ws.ws_server import WsServer

AGENT_PORT = 3030


async def _init_ws_server(app):
    app['ws_server'] = WsServer(app)


async def _stop_ws_server(app):
    await app['ws_server'].stop()


def run_server(
        env_settings=True,                          # get all settings from env variables
        ws_server=True,                             # start websocket server
        fs_exposed_dir='/app/',                     # root dir of all files put through handler /put_files/
        entrypoint_dir='/.utils/',                  # cwd of the script running after /run_script/
        request_max_body_size_mb=30,                # max body size of any request to agent
        startup_timeout=50,                         # /wait_server_started/ returns after waiting max seconds specified
        ports_to_wait=None,                         # ports to wait become *up* (default [3000])
        ports_to_wait_responding=None,              # ports to wait become *responding* (default [])
        proc_mem_limit_mb=512,                      # memory limit of process spawned through websocket
        wait_on_put=True,                           # wait server, before updating files with /put_files/
        interpreter_name='python',                  # interpreter used in /run_code/
        script_name='__test.py',                    # script name used in /run_code/
        entrypoint_path='/.utils/entrypoint.sh',    # script name to run on call /run_script/
        entypoint_logs_path='/.utils/logs',         # path to the file where entrypoint's logs will be stored
        proc_output_limit=2 ** 20,                  # max number of bytes in output
        proc_output_buffer_size=2 ** 16,            # max number of bytes in output per line
        output_processor_cls=BaseOutputProcessor,   # used to decode process's output into messages
        template_path=None,                         # path to get template to wrap code in (required)
        testlib_default_version='0.0.1',            # testlib default version used in python based trainers
        import_snippet_path='/.utils/testlib_import_snippet.py',    # path to get import snippet for py based trainers
):
    if ports_to_wait is None:
        ports_to_wait = [3000]
    if ports_to_wait_responding is None:
        ports_to_wait_responding = []

    logging.basicConfig(level=logging.DEBUG)

    REQUEST_MAX_BODY_SIZE_MB = _env_variables.REQUEST_MAX_BODY_SIZE_MB if env_settings else request_max_body_size_mb
    app = web.Application(client_max_size=REQUEST_MAX_BODY_SIZE_MB * 2 ** 20)

    if env_settings:
        init_settings_from_env(app)
    else:
        app['FS_EXPOSED_DIR'] = fs_exposed_dir
        app['ENTRYPOINT_DIR'] = entrypoint_dir
        app['REQUEST_MAX_BODY_SIZE_MB'] = request_max_body_size_mb
        app['STARTUP_TIMEOUT'] = startup_timeout
        app['PORTS_TO_WAIT'] = ports_to_wait
        app['PORTS_TO_WAIT_RESPONDING'] = ports_to_wait_responding
        app['PROC_MEM_LIMIT_MB'] = proc_mem_limit_mb
        app['WAIT_ON_PUT'] = wait_on_put
        app['INTERPRETER_NAME'] = interpreter_name
        app['SCRIPT_NAME'] = script_name
        app['ENTRYPOINT_PATH'] = entrypoint_path
        app['ENTRYPOINT_LOGS_PATH'] = entypoint_logs_path
        app['PROC_OUTPUT_LIMIT'] = proc_output_limit
        app['PROC_OUTPUT_BUFFER_SIZE'] = proc_output_buffer_size
        app['OUTPUT_PROCESSOR_CLS'] = output_processor_cls
        app['TEMPLATE_PATH'] = template_path
        app['TESTLIB_DEFAULT_VERSION'] = testlib_default_version
        app['IMPORT_SNIPPET_PATH'] = import_snippet_path

    app['filesystem'] = FileSystem(root_dir=app['FS_EXPOSED_DIR'])

    if ws_server:
        app.on_startup.append(_init_ws_server)
        app.on_shutdown.append(_stop_ws_server)

    setup_routes(app)

    web.run_app(
        app,
        host='0.0.0.0',
        port=AGENT_PORT,
    )



from typing import Any, Union, List

from marshmallow_dataclass import dataclass

# При исполнении пользовательского кода используется nbformat.v4
# https://github.com/jupyter/nbformat/blob/master/nbformat/v4/nbformat.v4.schema.json
# Чтобы отличать вывод тетрадки от пользовательского, stdout валидируется относительно
# формата вывода ExecutePreprocessor


@dataclass
class JupyterOutputExecuteResult:
    # execute_result
    output_type: Any
    execution_count: Any
    data: Any
    metadata: Any


@dataclass
class JupyterOutputDisplayData:
    # display_data
    output_type: Any
    data: Any
    metadata: Any


@dataclass
class JupyterOutputStream:
    # stream
    output_type: Any
    name: Any
    text: Any


@dataclass
class JupyterOutputError:
    # error
    output_type: Any
    ename: Any
    evalue: Any
    traceback: Any


@dataclass
class JupyterOutputs:
    outputs: List[
        Union[
            JupyterOutputError,
            JupyterOutputStream,
            JupyterOutputExecuteResult,
            JupyterOutputDisplayData,
        ]
    ]



import dataclasses
from typing import ClassVar, Text, Type

from marshmallow import EXCLUDE, Schema


class ObjectWithSchema:
    Schema: ClassVar[Type[Schema]]  # For the type checker

    def to_dict(self) -> dict:
        # TODO check errors
        return dataclasses.asdict(self)

    def to_json(self) -> Text:
        schema = self.Schema(unknown=EXCLUDE)
        return schema.dumps(self).data



import logging
from json import JSONDecodeError
from typing import Any, List

from aiohttp import web
from marshmallow import ValidationError

from serverhub_agent.types.handlers.common import build_drf_err
from serverhub_agent.types.handlers.put_files import PutFilesRequest
from serverhub_agent.utils.filesystem import FileSystem
from serverhub_agent.utils.runner import Runner
from serverhub_agent.utils.waiter import (PortConfig, ServerWaiter,
                                          WaitingProtocol)

logger = logging.getLogger(__name__)


async def put_files(request: web.Request) -> web.Response:  # TODO: i think we should use some auth here later
    """ Puts files to the specified dir inside the container """
    try:
        validated = PutFilesRequest.Schema().load(await request.json())  # type: (PutFilesRequest, Any)
    except (JSONDecodeError, ValidationError) as e:
        logger.warning(f'failed to deserialize request with error [{e}]')
        raise build_drf_err(code='', message=f"failed to deserialize request")

    # check that server is running, before we will update files
    if request.app['WAIT_ON_PUT']:
        server_ports_to_wait: List[PortConfig] = [
            PortConfig(port, WaitingProtocol.port_up.name)
            for port in request.app['PORTS_TO_WAIT'] if port
        ]
        server_ports_to_wait += [
            PortConfig(port, WaitingProtocol.port_responding.name)
            for port in request.app['PORTS_TO_WAIT_RESPONDING'] if port
        ]
        logger.info(f'server ports [{server_ports_to_wait}] {request.app["PORTS_TO_WAIT"]}')

        server_is_up = await ServerWaiter(server_ports_to_wait, 3).wait_for_ports()
        logger.warning(f"server ports: {server_ports_to_wait} is {'ready' if server_is_up else 'down'}")

        utils_ports_to_wait: List[PortConfig] = [
            PortConfig(port, WaitingProtocol.port_up.name)
            for port in [3001, 3002, 3003]  # TODO set only needed ports
        ]
        logger.info(f'utils ports [{utils_ports_to_wait}]')
        utils_is_up = await ServerWaiter(utils_ports_to_wait, 3).wait_for_ports()
        logger.warning(f"utils ports: {utils_ports_to_wait} is {'ready' if utils_is_up else 'down'}")

    fs = FileSystem(root_dir=request.app['FS_EXPOSED_DIR'])
    try:
        fs.create_files(validated.files)
    except ValueError as e:
        logger.warning(f'failed to store files with error [{e}]')
        raise build_drf_err(code='', message=f"messy file structure")

    if request.app['WAIT_ON_PUT'] and utils_is_up and not server_is_up:
        script_path = f"{request.app['ENTRYPOINT_DIR']}run_user_server.sh"
        logger.warning(f"all environment is up, except user server. rerun [{script_path}]")
        runner = Runner(
            working_dir=request.app['ENTRYPOINT_DIR'],
            script_path=script_path,
            logs_path=request.app['ENTRYPOINT_LOGS_PATH'],
        )
        await runner.run_script()

    return web.json_response({"status": "ok"})



import asyncio
from json import JSONDecodeError

from aiohttp import web
from marshmallow import ValidationError

from serverhub_agent.types.files import AstTests
from serverhub_agent.utils.filesystem import TempFileManager

TESTS_PATH = '/tests'


async def run_ast_test(request: web.Request) -> web.Response:
    body = await request.text()
    try:
        test_data = AstTests.Schema().loads(body)
    except (JSONDecodeError, ValidationError) as e:
        return web.json_response(data=f"Failed to parse ast test data: {e}", status=400)

    files_to_create = [test_data.answer, test_data.user, test_data.precode, test_data.test]

    with TempFileManager(directory=TESTS_PATH, files=files_to_create) as manager:
        ast_command = (
            f"python3 /testlibs/{test_data.testlib_version}.py "
            f"-u {test_data.user.name} "
            f"-a {test_data.answer.name} "
            f"-t {test_data.test.name} "
            f"-pre {test_data.precode.name}"
        )

        proc = await asyncio.create_subprocess_shell(
            f'cd {manager.directory} && exec {ast_command}',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        stdout, stderr = await proc.communicate()

    return web.json_response({
        "exit_code": proc.returncode,
        "stdout": stdout.decode(),
        "stderr": stderr.decode(),
    })



import setuptools

setuptools.setup(
    name="serverhub_agent",
    version="0.0.20",
    description="Runtime interface to running docker containers",
    url="https://github.yandex-team.ru/skills/serverhub-agent",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        "aiohttp==3.7.4.post0",
        "marshmallow-dataclass==7.5.2",
        "marshmallow_enum==1.5.1",
        "marshmallow_union==0.1.15",
        "psutil",
    ],
)



import logging

from aiohttp import web

from serverhub_agent.utils.runner import Runner

logger = logging.getLogger(__name__)


async def run_script(request: web.Request) -> web.Response:
    runner = Runner(
        working_dir=request.app["ENTRYPOINT_DIR"],
        script_path=request.app["ENTRYPOINT_PATH"],
        logs_path=request.app['ENTRYPOINT_LOGS_PATH'],
    )
    await runner.run_script()

    return web.json_response({"status": "ok"})



import logging

from aiohttp.web import Request, WebSocketResponse

from serverhub_agent.ws.ws_server import WsServer

logger = logging.getLogger(__name__)


async def ws_handler(request: Request) -> WebSocketResponse:
    ws_server: WsServer = request.app['ws_server']
    return await ws_server.connect(request)  # runs until socket is closed by the client



from aiohttp import web

from serverhub_agent.handlers.cpp_ws_handler import ws_handler
from serverhub_agent.handlers.ping import ping
from serverhub_agent.handlers.put_files import put_files
from serverhub_agent.handlers.run_ast_test import run_ast_test
from serverhub_agent.handlers.run_code import run_code
from serverhub_agent.handlers.run_script import run_script
from serverhub_agent.handlers.run_test import run_test
from serverhub_agent.handlers.wait_server_ready import wait_server_ready


def setup_routes(app: web.Application) -> None:
    app.router.add_post(f'/put_files/', put_files)
    app.router.add_post(f'/run_script/', run_script)
    app.router.add_get(f'/ping/', ping)
    app.router.add_get(f'/wait_server_ready/', wait_server_ready)
    app.router.add_get('/ws/', ws_handler)
    app.router.add_post('/run_code/', run_code)
    app.router.add_post('/run_test/', run_test)
    app.router.add_post('/run_ast_test/', run_ast_test)



import asyncio
import logging
from pathlib import Path
from typing import Text

logger = logging.getLogger(__name__)


class Runner:
    """ Runs entrypoint scripts for now, later it may run user's code and proxy logs etc. """
    def __init__(self, working_dir: Text, script_path: Text, logs_path: Text):
        self.working_dir = Path(working_dir)
        self.script_path = Path(script_path)
        self.logs_path = Path(logs_path)

    async def run_script(self):
        redirect = ''
        if self.logs_path:
            redirect = f'> {self.logs_path} 2>&1'

        proc = await asyncio.create_subprocess_shell(
            f'cd {self.working_dir} && /bin/bash {self.script_path} {redirect}',
        )
        logger.info(f'spawned entrypoint process with pid [{proc.pid}]')
        return



from typing import Callable, List, Text, Type

from aiohttp.web_exceptions import HTTPException
from aiohttp.web_response import Response
from marshmallow_dataclass import dataclass

from serverhub_agent.types.common import ObjectWithSchema


def get_all_subclasses(base_object: Type[HTTPException]) -> List[Type]:
    queue = base_object.__subclasses__()
    subclasses = []
    while len(queue):
        subclasses.append(queue.pop(0))
        for subclass in subclasses[-1].__subclasses__():
            queue.append(subclass)
    return subclasses


class HttpErrorFabric:

    base_exception: Type[HTTPException] = HTTPException

    def __init__(self):
        self.key_err_func = self._get_key_err_func()
        self.http_errors_map = self._build_err_map()

    def _get_key_err_func(self) -> Callable:
        return lambda err: getattr(err, 'status_code', -1) != -1

    def _build_err_map(self) -> dict:
        return {
            err.status_code: err
            for err in get_all_subclasses(self.base_exception)
            if self.key_err_func(err)
        }

    async def build(self, response: Response) -> HTTPException:

        err_class = self.class_by_status(response.status)
        return err_class(
            headers=response.headers,
            reason=response.reason,
            text=await response.text(),
        )

    def class_by_status(self, status: int):
        try:
            return self.http_errors_map[status]
        except KeyError:
            return self.base_exception


HttpErrorFabric = HttpErrorFabric()



@dataclass
class DRFError(ObjectWithSchema):
    code: Text
    message: Text


@dataclass
class DRFErrorResponse(ObjectWithSchema):
    errors: List[DRFError]


def build_err_string(code: Text, message: Text):
    response_data = DRFErrorResponse.Schema().dumps(
        DRFErrorResponse(
            errors=[DRFError(code=code, message=message)]),
    )
    return response_data


def build_drf_err(code: Text, message: Text, status: int = 400):
    return HttpErrorFabric.class_by_status(status)(
        content_type='application/json',
        text=build_err_string(code=code, message=message),
    )



import os
import shutil
from base64 import b64decode
from pathlib import Path
from typing import List, Text

from serverhub_agent.types.files import File, TestFile


class FileSystem:
    def __init__(self, root_dir: Text):
        self.root_dir: Path = Path(root_dir).absolute()
        self.tmp_files: List[Path] = []

    def create_file(self, file: File, is_tmp=False):
        if file.isDir:
            return

        users_file = Path(self.root_dir, Path(file.name)).resolve()
        users_path = users_file.parent

        # Raises ValueError if files is created in tricky pass, something like: ../../../../foo
        users_path.relative_to(self.root_dir)

        users_path.mkdir(parents=True, exist_ok=True)
        mode = 'w'
        content = file.content
        if file.isBin:
            mode = 'wb'
            content = b64decode(content)

        with users_file.open(mode) as f:
            f.write(content)

        if is_tmp:
            self.tmp_files.append(users_file)

    def create_files(self, files: List[File], is_tmp=False):
        for file in files:
            self.create_file(file, is_tmp)

    def remove_tmp(self):
        for f_path in self.tmp_files:
            try:
                f_path.unlink()
            except FileNotFoundError:
                ...
        self.tmp_files = []


class TempFileManager:
    directory = None
    files = None

    def create_file(self, temp_file: TestFile):
        file_path = f"{self.directory}/{temp_file.name}"

        with open(file_path, 'w') as f:
            f.writelines(temp_file.content)

    def __init__(self, *, files: List[TestFile], directory: str):
        self.files = files or []
        self.directory = directory

        if not os.path.exists(self.directory):
            os.makedirs(self.directory, exist_ok=True)

    def __enter__(self):
        for temp_file in self.files:
            self.create_file(temp_file)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        shutil.rmtree(self.directory)
        if exc_type is not None:
            return False
